/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.JBIException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.ObjectName;

import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.util.PropertiesUtil;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.SettableanyURI;
import org.ow2.petals.component.framework.mbean.MBeanHelper;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.component.framework.util.UtilFactory;

/**
 * A default bootstrap. This bootstrap class register an MBean object into the
 * JMX server that will be available only during the bootstrap phase. The
 * developer only needs to override methods {@code getMethodList()} to set the
 * list of method that he want to expose via JMX.
 * 
 * @author chamerling - EBM WebSourcing
 * @author rnaudin - EBM WebSourcing
 * 
 */
public class DefaultBootstrap extends JBIBootstrap {

    /**
     * The bootstrap MBean name
     */
    protected ObjectName mbeanName;

    /**
     * Returns the extension MBean name.
     * 
     * @return Returns the extension MBean name, except if the component is
     *         installed (ie. the extension MBean is not registered) in which
     *         case <code>null</code> is returned.
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    @Override
    public ObjectName getExtensionMBeanName() {

        final ObjectName objName;

        final MBeanServer server = this.installContext.getContext().getMBeanServer();
        if (server.isRegistered(this.mbeanName)) {
            objName = this.mbeanName;
        } else {
            objName = null;
        }

        return objName;
    }

    /**
     * Installs the component persisting the configuration done via the
     * extension Mbean.
     * 
     * @see javax.jbi.component.Bootstrap#onInstall()
     */
    @Override
    public void onInstall() throws JBIException {
        this.getLogger().info(getComponentMessageInfo());
        final File jbiXmlFile = new File(this.getInstallContext().getInstallRoot() + File.separator
                + Constants.Component.META_INF + File.separator + Constants.Component.JBI_XML);

        this.validateComponentSettableConfiguration();

        try {
            final FileOutputStream jbiOutputStream = new FileOutputStream(jbiXmlFile);
            try {
                JBIDescriptorBuilder.buildXmlJBIdescriptor(this.getJbiComponentConfiguration(),
                        jbiOutputStream);
            } finally {
                jbiOutputStream.close();
            }
        } catch (final IOException e) {
            throw new JBIException("Failed to persist the installation configuration : "
                    + e.getMessage());
        } catch (final CDKJBIDescriptorException e) {
            throw new JBIException("Failed to persist the installation configuration : "
                    + e.getMessage(), e);
        }
        try {
            this.processNativeWSDLResource(this.getInstallContext().getContext().getWorkspaceRoot()
                    + File.separator + Constants.Component.NATIVE_SERVICE_FOLDER);
        } catch (WSDLException e) {
            throw new JBIException("An error occurred while copying the native wsdl resource : "
                    + e.getMessage());
        }
        try {
            // @FIXME : The notifcation service is always started so the
            // resource are needed.
            this.processNotificationResource(this.getInstallContext().getContext()
                    .getWorkspaceRoot()
                    + File.separator
                    + Constants.Notification.NOTIFICATION_FOLDER
                    + File.separator
                    + Constants.Notification.NOTIFICATION_WSDL_FOLDER);
        } catch (WSDLException e) {
            throw new JBIException("An error occurred while copying the resource notification : "
                    + e.getMessage());
        }
    }

    /**
     * Clean resources allocated by the bootstrap
     * 
     * @see javax.jbi.component.Bootstrap#cleanUp()
     */
    @Override
    public void cleanUp() throws JBIException {

        MBeanServer server = installContext.getContext().getMBeanServer();
        if (server.isRegistered(mbeanName)) {
            try {
                server.unregisterMBean(mbeanName);
            } catch (InstanceNotFoundException e) {
                throw new JBIException("Error during bootstrap clean up : " + e.getMessage());
            } catch (MBeanRegistrationException e) {
                throw new JBIException("Error during bootstrap clean up" + e.getMessage());
            }
        }
    }

    /**
     * Initialize the bootstrap registering its Extension MBean into the JMX
     * server.
     * 
     * @see org.ow2.petals.component.framework.JBIBootstrap#doInit()
     */
    @Override
    protected void doInit() throws JBIException {

        this.mbeanName = this.installContext
                .getContext()
                .getMBeanNames()
                .createCustomComponentMBeanName(
                        "bootstrap_" + this.installContext.getComponentName());

        MBeanHelper.registerMBean(this.installContext.getContext().getMBeanServer(),
                this.mbeanName, new ExtensionMBean(this));
    }

    /**
     * Method to be overridden if the component implementation needs to expose
     * methods via JMX. This method is called by the dynamic MBean.
     * 
     * @return a list of method names to expose
     * 
     */
    protected List<String> getMethodList() {
        return new ArrayList<String>();
    }

    /**
     * Method to be overridden if the component implementation needs to expose
     * attributes via JMX. This method is called by the dynamic MBean.
     * 
     * @return a list of attributes names to expose
     * 
     */
    protected List<String> getAttributeList() {
        return new ArrayList<String>();
    }

    /**
     * Validate the Component settable configuration.
     * 
     * @throws PEtALSCDKException
     */
    private void validateComponentSettableConfiguration() throws JBIException {
        SettableanyURI propertiesFile = this.getJbiComponentConfiguration().getComponent()
                .getPropertiesFile();

        if (propertiesFile != null && !StringHelper.isNullOrEmpty(propertiesFile.getValue())) {
            try {
                PropertiesUtil.loadProperties(propertiesFile.getValue());
            } catch (PEtALSCDKException e) {
                throw new JBIException(e.getMessage());
            }
        }
    }

    /**
     * Copy all notification resources in the work path of the components.
     * 
     * @param workRootPath
     *            Path of the work directory of the component.
     * @throws WSDLException
     * @throws IOException
     */
    private void processNotificationResource(final String workRootPath) throws WSDLException {
        final File workspace = new File(workRootPath);
        if (!workspace.exists())
            workspace.mkdirs();
        final File producerWsdlFile = new File(workspace, Constants.Notification.PRODUCER_WSDL);
        if (producerWsdlFile != null && !producerWsdlFile.exists()) {
            final URL producerWsdlURL = this.getClass().getClassLoader()
                    .getResource(Constants.Notification.PRODUCER_WSDL);
            if (producerWsdlURL != null) {
                final Description wsdlDescription = UtilFactory.getWSDLUtil()
                        .createWsdlDescription(producerWsdlURL, true, false);
                UtilFactory.getWSDLUtil().convertDescriptionToFile(wsdlDescription,
                        producerWsdlFile);
            }
        }
        final File consumerWsdlFile = new File(workspace, Constants.Notification.CONSUMER_WSDL);
        if (consumerWsdlFile != null && !consumerWsdlFile.exists()) {
            final URL consumerWsdlURL = this.getClass().getClassLoader()
                    .getResource(Constants.Notification.CONSUMER_WSDL);
            if (consumerWsdlURL != null) {
                final Description wsdlDescription = UtilFactory.getWSDLUtil()
                        .createWsdlDescription(consumerWsdlURL, true, false);
                UtilFactory.getWSDLUtil().convertDescriptionToFile(wsdlDescription,
                        consumerWsdlFile);
            }
        }
        final String[] resourceNames = { "bw-2.wsdl", "rw-2.wsdl", "b-2.xsd", "bf-2.xsd",
                "r-2.xsd", "t-1.xsd", "ws-addr.xsd", "xml.xsd" };
        for (String resourceName : resourceNames) {
            final File resFile = new File(workspace, resourceName);
            if (resFile != null && !resFile.exists()) {
                final URL resourceUrl = this.getClass().getClassLoader().getResource(resourceName);
                if (resourceUrl == null) {
                    throw new WSDLException("Can not find resource : [" + resourceName + "]");
                }
                InputStream in;
                try {
                    in = resourceUrl.openStream();
                } catch (IOException e1) {
                    throw new WSDLException(e1);
                }
                if (in == null) {
                    throw new WSDLException("Can not open stream for the url : ["
                            + resourceUrl.toExternalForm() + "]");
                }
                try {
                    this.copyFile(in, resFile);
                } catch (IOException e) {
                    throw new WSDLException(e);
                }
            }
        }
    }

    /***
     * Copy the native WSDL resources in the work path of the components.
     * 
     * @param installRootPath
     *            Path of the work directory of the component.
     * @throws WSDLException
     */
    private void processNativeWSDLResource(final String workRootPath) throws WSDLException {
        final File workspace = new File(workRootPath);
        if (!workspace.exists())
            workspace.mkdirs();
        final File nativeWSDLFile = new File(workspace, Constants.Component.COMPONENT_WSDL);
        if (nativeWSDLFile != null && !nativeWSDLFile.exists()) {
            final URL nativeWsdlURL = this.getClass().getClassLoader()
                    .getResource(Constants.Component.COMPONENT_WSDL);
            if (nativeWsdlURL != null) {
                final Description wsdlDescription = UtilFactory.getWSDLUtil()
                        .createWsdlDescription(nativeWsdlURL, true);
                UtilFactory.getWSDLUtil().convertDescriptionToFile(wsdlDescription, nativeWSDLFile);
            }
        }
    }

    /**
     * Write the InputStream in the install directory with filename equals
     * destFileName.
     * 
     * @param is
     * @param destFileName
     * @throws IOException
     */
    private void copyFile(final InputStream is, final File destFile) throws IOException {
        final FileOutputStream fos = new FileOutputStream(destFile);
        final byte[] buf = new byte[1024];
        int len;
        try {
            while ((len = is.read(buf)) > 0) {
                fos.write(buf, 0, len);
            }
        } finally {
            is.close();
            fos.close();
        }
    }

    /**
     * Message to indicate component's name and component's version
     * 
     * @return Exemple : <br/>
     *         "Install component [Name : petals-bc-mail, Version : 3.1.4-SNAPSHOT]"
     */
    private String getComponentMessageInfo() {
        String componentVersion = "Version not set";
        final String clazz = this.getJbiComponentConfiguration().getComponent()
                .getJbiListenerClassName();
        try {
            Class<?> clazz2 = Class.forName(clazz, false, this.getClass().getClassLoader());
            if (clazz2 != null) {
                componentVersion = clazz2.getPackage().getImplementationVersion();
                if (StringHelper.isNullOrEmpty(componentVersion))
                    componentVersion = "Version not set";
            }
        } catch (ClassNotFoundException e1) {
        }
        final String result = "Install component [Name : "
                + this.getInstallContext().getComponentName() + ", Version : " + componentVersion
                + "]";
        return result;
    }
}
