/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.mbean;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.management.Attribute;
import javax.management.AttributeChangeNotification;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationEmitter;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.ReflectionException;

import org.ow2.petals.component.framework.ExtensionMBean;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Component;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositivestrictint;

/**
 * A runtime configuration MBean which sends notifications to subscribers. It is
 * used to change configuration values at runtime.
 * 
 * @author chamerling - EBM WebSourcing
 * @since 3.0
 * 
 */
public class RuntimeConfigurationNotifier implements DynamicMBean, NotificationEmitter {

    private final NotificationBroadcasterSupport broadcasterSupport;

    private ObjectName mbeanName;

    private final AtomicLong seqNumber;

    private final Logger logger;

    private final Jbi jbiComponentConfiguration;

    private final File jbiComponentFile;

    /**
     * 
     * @param workspaceRoot
     * @param properties
     * @throws JBIException
     */
    public RuntimeConfigurationNotifier(final Jbi jbiComponentConfiguration,
            final File jbiComponentFile, final Logger logger) {
        this.jbiComponentConfiguration = jbiComponentConfiguration;
        this.jbiComponentFile = jbiComponentFile;
        this.broadcasterSupport = new NotificationBroadcasterSupport();
        this.seqNumber = new AtomicLong(0);
        this.logger = logger;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getAttribute(java.lang.String)
     */
    public synchronized Object getAttribute(final String attribute)
            throws AttributeNotFoundException, MBeanException, ReflectionException {
        Object attributeObject = null;

        try {
            final Field field = Component.class.getDeclaredField(attribute);
            field.setAccessible(true);
            final Object gettableObject = field.get(this.jbiComponentConfiguration.getComponent());
            if (gettableObject != null) {
                String methodName = "getValue";
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    methodName = "isValue";
                }
                final Method method = gettableObject.getClass().getMethod(methodName);
                attributeObject = method.invoke(gettableObject);
            }

        } catch (final SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute + "' is not reachable");
        } catch (final NoSuchFieldException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute + "' does not exist");
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final NoSuchMethodException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e, e.getMessage());
        }

        return attributeObject;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getAttributes(java.lang.String[])
     */
    public synchronized AttributeList getAttributes(final String[] attributes) {
        final AttributeList list = new AttributeList();

        for (final String attribute : attributes) {
            try {
                list.add(new Attribute(attribute, this.getAttribute(attribute)));
            } catch (final AttributeNotFoundException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            } catch (final MBeanException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            } catch (final ReflectionException e) {
                this.logger.warning("Failed to get attribute '" + attribute
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            }
        }

        return list;
    }

    private Field findInheritedField(final Class<?> clazz, final String fieldName)
            throws NoSuchFieldException {

        Class<?> currentClazz = clazz;

        while (currentClazz != null) {
            try {
                return currentClazz.getDeclaredField(fieldName);
            } catch (final NoSuchFieldException e) {
                currentClazz = currentClazz.getSuperclass();
                if (currentClazz == null) {
                    // No more superclass
                    throw e;
                }
            }
        }
        // Statement unreachable
        return null;

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.DynamicMBean#setAttribute(javax.management.Attribute)
     */
    public synchronized void setAttribute(final Attribute attribute)
            throws AttributeNotFoundException, InvalidAttributeValueException, MBeanException,
            ReflectionException {

        try {
            final Field field = Component.class.getDeclaredField(attribute.getName());
            if (field.getType().equals(Runtimepositivestrictint.class)) {
                if (field.getName().equals(ExtensionMBean.PROCESSOR_MAX_POOL_SIZE_ATTR_NAME)
                        && ((Integer) attribute.getValue()).intValue() <= ((Integer) this
                                .getAttribute(ExtensionMBean.PROCESSOR_POOL_SIZE_ATTR_NAME)).intValue()) {
                    throw new InvalidAttributeValueException(
                            "Invalid value for attribute '"
                                    + attribute.getName()
                                    + "': The value must be upper or equals to the processor core pool size.");
                } else if (field.getName().equals(ExtensionMBean.PROCESSOR_POOL_SIZE_ATTR_NAME)
                        && ((Integer) attribute.getValue()).intValue() > ((Integer) this
                                .getAttribute(ExtensionMBean.PROCESSOR_MAX_POOL_SIZE_ATTR_NAME)).intValue()) {
                    throw new InvalidAttributeValueException(
                            "Invalid value for attribute '"
                                    + attribute.getName()
                                    + "': The value must be lower or equals to the processor max pool size.");
                } else if (((Integer) attribute.getValue()).intValue() <= 0) {
                        throw new InvalidAttributeValueException("Invalid value for attribute '"
                                + attribute.getName()
                                + "': The value must be upper or equals to 1.");
                }
            }
            field.setAccessible(true);
            final Object settableObject = field.get(this.jbiComponentConfiguration.getComponent());

            // get the previous value
            String methodName = "getValue";
            if (field.getType().getSimpleName().endsWith("boolean")) {
                methodName = "isValue";
            }
            Method method = settableObject.getClass().getMethod(methodName);
            final Object value = method.invoke(settableObject);

            // set the value
            final Field primitiveField = this.findInheritedField(field.getType(), "value");
            method = settableObject.getClass().getMethod("setValue",
                    new Class[] { primitiveField.getType() });
            method.invoke(settableObject, new Object[] { attribute.getValue() });

            // notify the listeners about the value modification
            this.saveAndNotifyListeners(attribute.getName() + "modified", attribute.getName(),
                    value, attribute.getValue());

        } catch (final SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute.getName()
                    + "' is not reachable");
        } catch (final NoSuchFieldException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute.getName()
                    + "' does not exist");
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e);
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e);
        } catch (final NoSuchMethodException e) {
            throw new ReflectionException(e);
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.DynamicMBean#setAttributes(javax.management.AttributeList
     * )
     */
    public synchronized AttributeList setAttributes(final AttributeList list) {
        final AttributeList attributeList = new AttributeList();

        for (final Object object : list) {
            final Attribute attribute = (Attribute) object;
            try {
                this.setAttribute(attribute);
            } catch (AttributeNotFoundException e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            } catch (InvalidAttributeValueException e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            } catch (MBeanException e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            } catch (ReflectionException e) {
                this.logger.warning("Failed to set attribute '" + attribute.getName()
                        + "' in the runtime configuration MBean : " + e.getClass().getSimpleName()
                        + " : " + e.getMessage());
            }
        }

        return attributeList;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getMBeanInfo()
     */
    public synchronized MBeanInfo getMBeanInfo() {
        // get the attributes
        final MBeanAttributeInfo[] attrs = this.getAttributesInfos();

        final MBeanNotificationInfo notif = new MBeanNotificationInfo(null,
                "Runtime MBean notifier", "Runtime configuration modification notifier");
        final MBeanNotificationInfo[] notifs = new MBeanNotificationInfo[] { notif };

        // create MBeanInfo with no constructors and no operations
        return new MBeanInfo(this.getClass().getName(), "Property Manager MBean", attrs, null,
                null, notifs);
    }

    public Object invoke(final String actionName, final Object[] params, final String[] signature)
            throws MBeanException, ReflectionException {
        // TODO Auto-generated method stub
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.NotificationBroadcaster#getNotificationInfo()
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] { new MBeanNotificationInfo(
                new String[] { AttributeChangeNotification.ATTRIBUTE_CHANGE },
                AttributeChangeNotification.class.getName(),
                "Change runtime configuration attribute") };
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#addNotificationListener(javax
     * .management.NotificationListener, javax.management.NotificationFilter,
     * java.lang.Object)
     */
    public void addNotificationListener(final NotificationListener listener,
            final NotificationFilter filter, final Object handback) {
        this.broadcasterSupport.addNotificationListener(listener, filter, handback);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#removeNotificationListener(javax
     * .management.NotificationListener)
     */
    public void removeNotificationListener(final NotificationListener listener)
            throws ListenerNotFoundException {
        this.broadcasterSupport.removeNotificationListener(listener);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationEmitter#removeNotificationListener(javax
     * .management.NotificationListener, javax.management.NotificationFilter,
     * java.lang.Object)
     */
    public void removeNotificationListener(final NotificationListener listener,
            final NotificationFilter filter, final Object handback)
            throws ListenerNotFoundException {
        this.broadcasterSupport.removeNotificationListener(listener, filter, handback);
    }

    /**
     * Get the MBean Name
     * 
     * @return
     */
    public ObjectName getMBeanName() {
        return this.mbeanName;
    }

    /**
     * 
     * @param mbeanName
     */
    public void setMBeanName(final ObjectName mbeanName) {
        this.mbeanName = mbeanName;
    }

    /**
     * Dynamically get attributes for the MBean. The attributes are the ones
     * from the properties file.
     * 
     * @return
     */
    protected MBeanAttributeInfo[] getAttributesInfos() {
        MBeanAttributeInfo[] attrs = null;
        final Field[] fields = Component.class.getDeclaredFields();
        final List<Field> settableFields = new ArrayList<Field>();

        try {
            // get the settable fields
            for (final Field field : fields) {
                if (field.getType().getSimpleName().startsWith("Runtime")) {
                    field.setAccessible(true);
                    if (field.get(this.jbiComponentConfiguration.getComponent()) != null) {
                        settableFields.add(field);
                    }
                }
            }

            // build the MBeanAttributeInfo
            attrs = new MBeanAttributeInfo[settableFields.size()];
            for (final Field field : settableFields) {
                final Object settableObject = field.get(this.jbiComponentConfiguration
                        .getComponent());
                Method getter;
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    getter = settableObject.getClass().getMethod("isValue");
                } else {
                    getter = settableObject.getClass().getMethod("getValue");
                }
                final Field primitiveField = this.findInheritedField(field.getType(), "value");
                final Method setter = settableObject.getClass().getMethod("setValue",
                        new Class[] { primitiveField.getType() });

                attrs[settableFields.indexOf(field)] = new MBeanAttributeInfo(field.getName(),
                        "Property " + field.getName(), getter, setter);

            }

        } catch (final IllegalArgumentException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final IllegalAccessException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final SecurityException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final NoSuchMethodException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final IntrospectionException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final NoSuchFieldException e) {
            this.logger.warning("Failed to initialize fields of the runtime configuration MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        }

        return attrs;
    }

    /**
     * FIXME : New value to be saved in the properties file.
     * 
     * @param msg
     * @param attrName
     * @param oldVal
     * @param newVal
     */
    private void saveAndNotifyListeners(final String msg, final String attrName,
            final Object oldVal, final Object newVal) {
        OutputStream jbiComponentFileAsOutputStream = null;
        try {
            jbiComponentFileAsOutputStream = new FileOutputStream(this.jbiComponentFile);
            JBIDescriptorBuilder.buildXmlJBIdescriptor(this.jbiComponentConfiguration,
                    jbiComponentFileAsOutputStream);
        } catch (final FileNotFoundException e) {
            this.logger
                    .warning("Failed to pesist modification of the runtime configuration MBean : "
                            + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final CDKJBIDescriptorException e) {
            this.logger
                    .warning("Failed to pesist modification of the runtime configuration MBean : "
                            + e.getClass().getSimpleName() + " : " + e.getMessage());
        } finally {
            try {
                if (jbiComponentFileAsOutputStream != null) {
                    jbiComponentFileAsOutputStream.close();
                }
            } catch (final IOException ex) {
                this.logger.warning("Can't close jbi component file output stream : "
                        + ex.getMessage());
            }
        }

        final Notification notif = new AttributeChangeNotification(this, this.seqNumber
                .getAndIncrement(), System.currentTimeMillis(), msg, attrName, String.class
                .getName(), oldVal, newVal);
        this.broadcasterSupport.sendNotification(notif);
    }
}
