/**
 * PETALS: PETALS Services Platform Copyright (C) 2009 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 */

package org.ow2.petals.component.framework.notification;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.notification.StatusListener;
import org.ow2.petals.component.framework.util.CustomizableThreadFactory;

import com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Notify;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Subscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.SubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Unsubscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.UnsubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.DestroyRegistration;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.DestroyRegistrationResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisher;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisherResponse;
import com.ebmwebsourcing.wsstar.notification.extension.api.SOAParameterType;

/**
 * Default {@link AsyncSender} implementation. Launch tasks in executor and pass
 * result to the listener.
 * 
 * @author chamerling - eBM WebSourcing
 * 
 */
public class AsyncSender implements org.ow2.petals.component.framework.api.notification.AsyncSender {

    private WrappedSender sender;

    private StatusListener listener;

    private ExecutorService executor;

    /**
     * 
     * @param sender
     * @param listener
     */
    public AsyncSender(WrappedSender sender, StatusListener listener) {
        this.sender = sender;
        this.listener = listener;

        final String componentName = this.sender.getComponentContext().getComponentName();
        this.executor = Executors.newCachedThreadPool(new CustomizableThreadFactory(componentName
                + " - Async Sender Thread #"));
    }

    public StatusListener getStatusListener() {
        return this.listener;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void send(final DestroyRegistration destroyRegistration) throws PEtALSCDKException {
        this.executor.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    DestroyRegistrationResponse response = sender.send(destroyRegistration);
                    if (getStatusListener() != null) {
                        getStatusListener().onUnregistered(response);
                    }
                } catch (Exception e) {
                    if (getStatusListener() != null) {
                        getStatusListener().onException(e);
                    }
                }
            }
        });

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void send(final DestroyRegistration destroyRegistration,
            final SOAParameterType soaParameterType) throws PEtALSCDKException {
        this.executor.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    DestroyRegistrationResponse response = sender.send(destroyRegistration,
                            soaParameterType);
                    if (getStatusListener() != null) {
                        getStatusListener().onUnregistered(response);
                    }
                } catch (Exception e) {
                    if (getStatusListener() != null) {
                        getStatusListener().onException(e);
                    }
                }
            }
        });
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void send(final RegisterPublisher registerPublisher) throws PEtALSCDKException {
        this.executor.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    RegisterPublisherResponse response = sender.send(registerPublisher);
                    if (getStatusListener() != null) {
                        getStatusListener().onRegistered(response);
                    }
                } catch (Exception e) {
                    if (getStatusListener() != null) {
                        getStatusListener().onException(e);
                    }
                }
            }
        });

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void send(final Subscribe subscribe, final QName targetServName,
            final String targetEpName) {
        this.executor.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    SubscribeResponse response = sender.send(subscribe, targetServName,
                            targetEpName);
                    if (getStatusListener() != null) {
                        getStatusListener().onSubscribed(response);
                    }
                } catch (Exception e) {
                    if (getStatusListener() != null) {
                        getStatusListener().onException(e);
                    }
                }
            }
        });
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void send(final Subscribe subscribe) throws PEtALSCDKException {
        this.executor.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    SubscribeResponse response = sender.send(subscribe);
                    if (getStatusListener() != null) {
                        getStatusListener().onSubscribed(response);
                    }
                } catch (Exception e) {
                    if (getStatusListener() != null) {
                        getStatusListener().onException(e);
                    }
                }
            }
        });
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void send(final Unsubscribe unsubscribe) throws PEtALSCDKException {
        this.executor.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    UnsubscribeResponse response = sender.send(unsubscribe);
                    if (getStatusListener() != null) {
                        getStatusListener().onUnsubscribed(response);
                    }
                } catch (Exception e) {
                    if (getStatusListener() != null) {
                        getStatusListener().onException(e);
                    }
                }
            }
        });
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void send(final Notify notify, final EndpointReferenceType consumerReference)
            throws PEtALSCDKException {
        this.executor.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    sender.send(notify, consumerReference);
                } catch (Exception e) {
                    if (getStatusListener() != null) {
                        getStatusListener().onException(e);
                    }
                }
            }
        });
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void abort() throws PEtALSCDKException {
        if (this.executor != null) {
            this.executor.shutdownNow();
        }
    }

    public boolean isStopped() {
        if (this.executor != null) {
            return this.executor.isShutdown();
        } else {
            return true;
        }
    }

}
