/**
 * PETALS: PETALS Services Platform Copyright (C) 2009 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 */

package org.ow2.petals.component.framework.notification;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.io.FileUtils;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.notification.LifeCyle;
import org.ow2.petals.component.framework.util.CustomizableThreadFactory;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.w3c.dom.Document;

import com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Notify;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisherResponse;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationReader;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationWriter;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;

/**
 * Service used to store notifications (which can not be sent. The notifications
 * will be sent ASAP by a processor), publisher registration responses ...
 * 
 * @author chamerling - eBM WebSourcing
 * 
 */
public class NotificationStorageService implements LifeCyle {

    static final String TRASH = "trash";

    static final String PERSIST = "persist";

    static final String PUB_REG_RESPONSE = "PubRegResponse";

    static final String POISON = "poison";

    protected static final long RESEND_DELAY = 120;

    private STATE state = STATE.UNKNOWN;

    private final String rootFolder;

    /**
     * Base folder for notification related stuff
     */
    private File folder;

    /**
     * Base folder for notification related stuff
     */
    private File persistFolder;

    /**
     * A trash folder used to move notification which can not be deleted...
     */
    private File trashFolder;

    /**
     * A folder to persist publisher registration responses
     */
    private File pubRegResponseFolder;

    /**
     * The publisher registration response file
     */
    private File pubRegResponseFile;

    private final Logger logger;

    private final AtomicLong fileCounter;

    private final org.ow2.petals.component.framework.api.notification.Sender sender;

    private NotificationContext notificationContext;

    private ScheduledExecutorService executor;

    /**
     * 
     */
    public NotificationStorageService(final String folder,
            final org.ow2.petals.component.framework.api.notification.Sender sender,
            NotificationContext notificationContext, final Logger logger) {
        this.rootFolder = folder;
        this.logger = logger;
        this.sender = sender;
        this.notificationContext = notificationContext;
        this.fileCounter = new AtomicLong();
    }

    /**
     * Store a notification
     * 
     * @throws PEtALSCDKException
     */
    void store(com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Notify notify)
            throws PEtALSCDKException {

        this.logger.info("Storing notification");
        try {
            final Document document = WSNotificationWriter.getInstance().writeNotify(notify);
            final String id = this.getFileName();
            final File persistFile = new File(this.persistFolder, id);
            if (persistFile.exists()) {
                persistFile.createNewFile();
            }
            XMLUtil.writeDocument(document, new FileOutputStream(persistFile));
        } catch (final Exception e) {
            if (this.logger.isLoggable(Level.WARNING)) {
                this.logger.log(Level.WARNING, "Failed to store notification.", e);
            }
        }

        // restart storage service if not already started
        this.start();
    }

    /**
     * method used to persist {@link RegisterPublisherResponse} java object to a
     * file
     * 
     * @param registerPublisherResponse
     *            the {@link RegisterPublisherResponse} java Object to persist
     * @throws WSNotificationException
     */
    public void store(RegisterPublisherResponse registerPublisherResponse) throws Exception {
        Document regPubRespAsDom = WSNotificationWriter.getInstance()
                .writeRegisterPublisherResponse(registerPublisherResponse);
        XMLUtil.writeDocument(regPubRespAsDom, new FileOutputStream(pubRegResponseFile));
    }

    /**
     * method used to remove persisted {@link RegisterPublisherResponse} file
     * 
     * @throws WSNotificationException
     */
    public void destroyPubRegResponseFile() throws Exception {
        if (pubRegResponseFile.exists()) {
            pubRegResponseFile.delete();
        }
    }

    public RegisterPublisherResponse restorePublisherRegistration() throws WSNotificationException {
        RegisterPublisherResponse result = null;
        if (this.pubRegResponseFolder.exists()) {
            File[] persistedRegPubRespFiles = this.pubRegResponseFolder.getAbsoluteFile()
                    .listFiles();
            if (persistedRegPubRespFiles != null && persistedRegPubRespFiles.length == 1) {
                try {
                    result = WSNotificationReader.getInstance().readRegisterPublisherResponse(
                            XMLUtil.loadDocument(new FileInputStream(persistedRegPubRespFiles[0])));
                } catch (Exception e) {
                    throw new WSNotificationException(e);
                }
            }
        }
        return result;
    }

    /**
     * @param notify
     * @return
     */
    private String getFileName() {
        // get a counter and the date (can be useful some day...)
        return this.fileCounter.incrementAndGet() + "-" + System.currentTimeMillis();
    }

    /**
     * 
     */
    public void init() {
        if (this.folder == null) {
            this.logger.fine("Creating notification folder");
            this.folder = new File(this.rootFolder, Constants.Notification.NOTIFICATION_FOLDER);
            this.folder.mkdirs();
        }

        if (this.persistFolder == null) {
            this.logger.fine("Creating notification persistance folder");
            this.persistFolder = new File(this.folder, PERSIST);
            this.persistFolder.mkdirs();
        }

        if (this.trashFolder == null) {
            this.logger.fine("Creating notification persistance backup folder");
            this.trashFolder = new File(this.folder, TRASH);
            this.trashFolder.mkdirs();
        }

        if (this.pubRegResponseFolder == null) {
            this.logger.fine("Creating register publisher response persistance backup folder");
            this.pubRegResponseFolder = new File(this.folder, PUB_REG_RESPONSE);
            this.pubRegResponseFolder.mkdirs();
        }

        pubRegResponseFile = new File(pubRegResponseFolder, PUB_REG_RESPONSE + ".xml");

        // TODO : load all the persisted notification from the folder in the
        // process queue so that the processor will send all the remaining files
        // on restart

        this.state = STATE.INIT;
    }

    /**
     * @param f
     * @throws IOException
     */
    protected void moveToTrash(final File f) throws IOException {
        FileUtils.moveFile(f, new File(this.trashFolder, f.getName()));
    }

    /**
     * 
     */
    public synchronized void start() {
        if (!STATE.STARTED.equals(state)) {
            // Run a scheduler that try to send persisted notifications
            // periodically
            this.runPersistedNotifProcessor();
            this.state = STATE.STARTED;
        }
    }

    private void runPersistedNotifProcessor() {

        final String componentName = this.sender.getComponentContext().getComponentName();
        this.executor = new ScheduledThreadPoolExecutor(1, new CustomizableThreadFactory(
                componentName + " - Persisted notif sender Thread#"));

        Runnable runnable = new Runnable() {

            @Override
            public void run() {

                // Add all persist file to queue
                File[] persistedNotifs = persistFolder.listFiles();
                if (persistedNotifs != null && persistedNotifs.length > 0) {

                    for (File file : persistedNotifs) {
                        try {
                            String fileName = file.getName();
                            NotificationStorageService.this.logger
                                    .fine("Processing the file with ID = " + fileName);

                            if (NotificationStorageService.this.notificationContext.isRegistered()) {
                                if (file.exists()) {
                                    final Document doc = XMLUtil.loadDocument(new FileInputStream(
                                            file));
                                    final Notify notify = WSNotificationReader.getInstance()
                                            .readNotify(doc);
                                    NotificationStorageService.this.sender.send(notify,
                                            NotificationStorageService.this.getConsumerReference());
                                    // if we are here this is a success,
                                    // so delete the file...
                                    final boolean deleted = FileUtils.deleteQuietly(file);
                                    if (!deleted) {
                                        NotificationStorageService.this.moveToTrash(file);
                                    }
                                } else {
                                    NotificationStorageService.this.logger.fine("File '"
                                            + file.getAbsolutePath() + "' does not exists for ID "
                                            + fileName);
                                }
                            } else {
                                NotificationStorageService.this.logger
                                        .fine("Can't process persisted notifications "
                                                + "if the publisher isn't registered. Retry later");
                            }
                        } catch (Exception e) {
                            if (NotificationStorageService.this.logger.isLoggable(Level.FINE)) {
                                NotificationStorageService.this.logger.log(Level.FINE,
                                        "An error occured during persisted "
                                                + "notification processing", e);
                            }
                        }

                    }
                } else {
                    NotificationStorageService.this.logger
                            .fine("No persisted notification to process. "
                                    + "Stop persisted notification processor.");
                    NotificationStorageService.this.stop();
                }
            }
        };
        NotificationStorageService.this.logger
                .info("Storage Service is now running to process persisted notifications");
        this.executor.scheduleAtFixedRate(runnable, RESEND_DELAY, RESEND_DELAY, TimeUnit.SECONDS);

    }

    /**
     * @return
     */
    protected EndpointReferenceType getConsumerReference() {
        return notificationContext.getConsumerReference();
    }

    /**
     * TODO synchronize stop method on a lock different from the start lock to
     * avoid deadlock
     */
    public void stop() {
        if (!STATE.STOPPED.equals(state)) {
            if (executor != null) {
                this.executor.shutdownNow();
                this.executor = null;
            }

            // clear the trashed files...
            try {
                FileUtils.cleanDirectory(trashFolder);
            } catch (IOException e) {
                this.logger.warning("Notification trash folder can't be cleaned. "
                        + "Will be cleaned later");
            }

            this.state = STATE.STOPPED;
        }
    }
}
