/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.process;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.management.AttributeChangeNotification;
import javax.management.Notification;
import javax.management.NotificationListener;

import org.apache.commons.pool.PoolUtils;
import org.apache.commons.pool.PoolableObjectFactory;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.component.framework.util.StringHelper;

/**
 * The JBI processor manager is in charge of JBI Message processor threads and
 * instances.
 * 
 * @author chamerling - EBM WebSourcing
 * @since 3.0
 * 
 */
public class JBIProcessorManager {
    
    private static final int DEFAULT_PROCESSOR_POOL_SIZE_MAX = 50;
    
    private static final int DEFAULT_PROCESSOR_POOL_SIZE_KEEP_ALIVE_TIME = 60;

    private ThreadPoolExecutor processorPool = null;

    protected GenericObjectPool objectPool = null;

    protected AbstractComponent component;

    private final RuntimeConfigurationNotifier runtimeConfiguration;

    protected Logger logger;

    /**
     * Inner Notification Listener class.
     * 
     * @author rnaudin
     */
    private class RuntimeConfigurationListener implements NotificationListener {
        /*
         * (non-Javadoc)
         * 
         * @see
         * javax.management.NotificationListener#handleNotification(javax.management
         * .Notification, java.lang.Object)
         */
        public void handleNotification(final Notification notification, final Object obj) {
            if (notification instanceof AttributeChangeNotification) {
                final AttributeChangeNotification attrNotif = (AttributeChangeNotification) notification;
                JBIProcessorManager.this.logger.log(Level.FINE,
                        "Notification received for attribute : " + attrNotif.getAttributeName());
                if (attrNotif.getAttributeName().equals("processorPoolSize")) {
                    final Integer newVal = (Integer) attrNotif.getNewValue();
                    JBIProcessorManager.this.setCorePoolSize(newVal.intValue());
                } else if (attrNotif.getAttributeName().equals("processorMaxPoolSize")) {
                    final Integer newVal = (Integer) attrNotif.getNewValue();
                    JBIProcessorManager.this.setMaxCorePoolSize(newVal.intValue());
                } else if (attrNotif.getAttributeName().equals("processorKeepAliveTime")) {
                    final Integer newVal = (Integer) attrNotif.getNewValue();
                    JBIProcessorManager.this.setPoolKeepAliveTime(newVal.intValue());
                }
            }
        }
    }

    /**
     * Inner Thread Factory class.
     * 
     * @author rnaudin
     */
    private class JBIProcessorThreadFactory implements ThreadFactory {

        final AtomicInteger threadNumber = new AtomicInteger(1);

        private final String componentName;

        /**
         * Create a new instance of {@link JBIListenerThreadFactory}
         * 
         * @param componentName
         * @param logger
         */
        public JBIProcessorThreadFactory() {
            this.componentName = JBIProcessorManager.this.component.getComponentConfiguration()
                    .getIdentification().getName();
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.util.concurrent.ThreadFactory#newThread(java.lang.Runnable)
         */
        public Thread newThread(final Runnable r) {
            final String name = this.componentName + " - MessageExchange Processor thread #"
                    + Integer.toString(this.threadNumber.getAndIncrement());
            JBIProcessorManager.this.logger.log(Level.FINE, "Creating new JBI Processor Thread : "
                    + name);
            return new Thread(Thread.currentThread().getThreadGroup(), r, name);
        }
    }

    /**
     * Inner Thread Pool executor class.
     * 
     * @author rnaudin
     * 
     */
    private class MessageExchangeProcessorThreadPoolExecutor extends ThreadPoolExecutor {

        public MessageExchangeProcessorThreadPoolExecutor(final int corePoolSize,
                final int maximumPoolSize, final long keepAliveTime, final TimeUnit unit,
                final BlockingQueue<Runnable> workQueue, final ThreadFactory threadFactory) {
            super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue, threadFactory,
                    new ThreadPoolExecutor.DiscardOldestPolicy());
        }

        @Override
        protected void afterExecute(final Runnable r, final Throwable t) {
            // Clean the Message Exchange Processor
            ((MessageExchangeProcessor) r).clean();
            try {
                JBIProcessorManager.this.objectPool.returnObject(r);
            } catch (final Exception e) {
                JBIProcessorManager.this.logger
                        .log(Level.SEVERE,
                                "Failed to return a Message Exchange Processor instance to the Object pool",
                                e);
            }
            super.afterExecute(r, t);
        }

    }

    /**
     * Inner Poolable Object Factory.
     * 
     * @author rnaudin
     * 
     */
    private class MessageExchangeProcessorObjectFactory implements PoolableObjectFactory {

        public void activateObject(final Object arg0) throws Exception {
            // do nothing

        }

        public void destroyObject(final Object arg0) throws Exception {
            // do nothing

        }

        public Object makeObject() throws Exception {
            final AbstractJBIListener jbiListener = this.createAndInitJBIListener();

            final MessageExchangeProcessor jbiProcessor = new MessageExchangeProcessor(
                    JBIProcessorManager.this.component, jbiListener,
                    JBIProcessorManager.this.logger);
            return jbiProcessor;
        }

        public void passivateObject(final Object arg0) throws Exception {
            // do nothing

        }

        public boolean validateObject(final Object arg0) {
            return true;
        }

        /**
         * Instantiate and initialize a JBI listener.s This JBI Listener is the
         * one which is defined by the component developer.
         * 
         * @return
         * @throws PEtALSCDKException
         */
        private AbstractJBIListener createAndInitJBIListener() throws PEtALSCDKException {
            AbstractJBIListener jbiListener = null;
            final String jbiListenerClassNameValue = JBIProcessorManager.this.component
                    .getComponentConfiguration().getJbiListenerClassName();
            final String jbiListenerClassName = jbiListenerClassNameValue.replaceAll("\\s", "");

            if (!StringHelper.isNullOrEmpty(jbiListenerClassName)) {
                try {
                    final ClassLoader componentClassLoader = this.getClass().getClassLoader();
                    final Class<?> clazz = componentClassLoader.loadClass(jbiListenerClassName);
                    jbiListener = (AbstractJBIListener) clazz.newInstance();
                } catch (final InstantiationException e) {
                    throw new PEtALSCDKException("Can't instanciate JBIListener : "
                            + jbiListenerClassName, e);
                } catch (final IllegalAccessException e) {
                    throw new PEtALSCDKException("Can't instanciate JBIListener : "
                            + jbiListenerClassName, e);
                } catch (final ClassNotFoundException e) {
                    throw new PEtALSCDKException("Can't find JBIListener class : "
                            + jbiListenerClassName, e);
                }
                jbiListener.init(JBIProcessorManager.this.component);
                // component implementor specific init
                jbiListener.init();
            } else {
                throw new PEtALSCDKException("JBIListener class name can not be empty");
            }
            return jbiListener;
        }

    }

    /**
     * Creates a new instance of {@link JBIProcessorManager}
     * 
     * @param component
     * @param configuration
     * @param logger
     */
    public JBIProcessorManager(final AbstractComponent component,
            final RuntimeConfigurationNotifier runtimeConfiguration, final Logger logger) {
        this.component = component;
        this.runtimeConfiguration = runtimeConfiguration;
        this.logger = logger;

        this.runtimeConfiguration.addNotificationListener(new RuntimeConfigurationListener(), null,
                null);
    }

    /**
     * Start the Processor Thread Pool.
     */
    public void start() throws JBIException {
        this.logger.fine("Creating JBI Processor pool:");
        
        final int coreProcessorPoolSize = this.component.getComponentConfiguration().getProcessorPoolSize().getValue();
        this.logger.fine("\t- core pool size: " + coreProcessorPoolSize);

        final int maxProcessorPoolSize;
        if (this.component.getComponentConfiguration().getProcessorMaxPoolSize() != null) {
            maxProcessorPoolSize = this.component.getComponentConfiguration()
                    .getProcessorMaxPoolSize().getValue();
        }
        else {
            maxProcessorPoolSize = DEFAULT_PROCESSOR_POOL_SIZE_MAX;
        }
        this.logger.fine("\t- max core pool size: " + maxProcessorPoolSize);
        
        final int processorPoolKeepAlive;
        if (this.component.getComponentConfiguration().getProcessorKeepAliveTime() != null) {
            processorPoolKeepAlive = this.component.getComponentConfiguration().getProcessorKeepAliveTime().getValue();
        }
        else {
            processorPoolKeepAlive = DEFAULT_PROCESSOR_POOL_SIZE_KEEP_ALIVE_TIME;
        }
        this.logger.fine("\t- keep alive time: " + processorPoolKeepAlive);

        this.objectPool = new GenericObjectPool(new MessageExchangeProcessorObjectFactory(),
                maxProcessorPoolSize);
        // initialize the pool with N objects
        try {
            PoolUtils.prefill(this.objectPool, coreProcessorPoolSize);
        } catch (final IllegalArgumentException e) {
            throw new JBIException(e);
        } catch (final Exception e) {
            throw new JBIException(e);
        }

        this.processorPool = new MessageExchangeProcessorThreadPoolExecutor(
                coreProcessorPoolSize,
                maxProcessorPoolSize,
                processorPoolKeepAlive, TimeUnit.SECONDS,
                new SynchronousQueue<Runnable>(),
                new JBIProcessorThreadFactory());

        this.processorPool.prestartAllCoreThreads();
    }

    /**
     * Stop the Processor Thread Pool.
     * 
     */
    public void stop() {
        // try to let the executing tasks finish
        this.processorPool.shutdown();
        try {
            long processorStopMaxWait = 15000;
            if (this.component.getComponentConfiguration().getProcessorStopMaxWait() != null) {
                processorStopMaxWait = this.component.getComponentConfiguration()
                        .getProcessorStopMaxWait().getValue();
            }
            this.processorPool.awaitTermination(processorStopMaxWait, TimeUnit.MILLISECONDS);
        } catch (final InterruptedException e) {
            this.logger.log(Level.SEVERE, "Workers pool interrupted : " + e.getMessage());
        }
        // interrupt the remaining executing tasks
        this.processorPool.shutdownNow();

        try {
            this.objectPool.close();
        } catch (final Exception e) {
            this.logger.log(Level.WARNING,
                    "Failed to close the Messsage Exchange Processor Object pool properly", e);
        }
    }

    /**
     * Get the Processor Thread Pool.
     */
    public ThreadPoolExecutor getThreadPool() {
        return this.processorPool;
    }

    /**
     * Get the Processor Object Pool.
     */
    public GenericObjectPool getObjectPool() {
        return this.objectPool;
    }

    /**
     * Reset the core Pool size.
     * 
     * @param size
     */
    private void setPoolKeepAliveTime(final int time) {
        this.logger.info("Reset the processor thread pool keep alive time to " + time);
        try {
            this.processorPool.setKeepAliveTime(time, TimeUnit.SECONDS);
        } catch (final IllegalArgumentException e) {
            this.logger
                    .warning("Invalid value used (less than or equal to zero) to configure the processor thread pool keep alive time: "
                            + time);
        }
    }

    /**
     * Reset the max core Pool size.
     * 
     * @param size
     */
    private void setMaxCorePoolSize(final int size) {
        this.logger.info("Reset the maximum processor thread pool size to " + size);
        try {
            this.processorPool.setMaximumPoolSize(size);
        } catch (final IllegalArgumentException e) {
            this.logger
                    .warning("Invalid value used (less than or equal to zero, or less than the core pool size) to configure the processor thread pool max size: "
                            + size);
        }
    }

    /**
     * Reset the core Pool size.
     * 
     * @param size
     */
    private void setCorePoolSize(final int size) {
        this.logger.info("Reset the processor thread pool size to " + size);
        try {
            this.processorPool.setCorePoolSize(size);
        } catch (final IllegalArgumentException e) {
            this.logger
                    .warning("Invalid value used (less than or equal to zero) to configure the processor thread pool core size: "
                            + size);
        }
    }
}
