/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.process;

import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.commons.util.IDGenerator;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.Message.InterceptorPhase;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.exception.SOAP11FaultServerException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.ExchangeProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.IgnoreStatus;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageParam;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.Param;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.notification.notify.AbstractProducerComponentNotify;
import org.ow2.petals.component.framework.notification.notify.FaultProducerComponentNotify;
import org.ow2.petals.component.framework.notification.notify.InProducerComponentNotify;
import org.ow2.petals.component.framework.notification.notify.OutProducerComponentNotify;
import org.ow2.petals.component.framework.notification.notify.StatusProducerComponentNotify;
import org.ow2.petals.component.framework.process.async.AsyncContext;
import org.ow2.petals.component.framework.process.async.AsyncMessageManager;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.component.framework.util.UtilFactory;

/**
 * Processor for received MessageExchanges. This processor can be overridden to
 * define a specific strategy to accept messages from the DeliveryChannel. See
 * <code>acceptMessageExchange()</code>. <br>
 * Override the <code>createMessageExchangeProcessor</code> method in the
 * <code>Component</code> class to set your processor.
 * 
 * @author alouis,ofabre,chamerling - EBM Websourcing
 * @since 2.0
 * 
 */
public class MessageExchangeProcessor implements Runnable {

    /**
     * The component logger;
     */
    private final AbstractComponent component;

    /**
     * The component logger
     */
    private final Logger logger;

    /**
     * The component delivery channel
     */
    private final DeliveryChannel deliveryChannel;

    /**
     * The component transaction manager
     */
    private final Object transactionManager;

    /**
     * The component implementation of the message processor
     */
    private final AbstractJBIListener jbiListener;

    /**
     * The component service unit manager
     */
    private final AbstractServiceUnitManager suManager;

    /**
     * The component asynchronous message manager
     */
    private final AsyncMessageManager asyncManager;

    /**
     * The component name
     */
    private final String componentName;

    /**
     * The component ignore status flag
     */
    private final IgnoreStatus ignoreStatus;

    /**
     * The exchange, during one message processing
     */
    private ExchangeImpl exchange = null;

    /**
     * Creates a processor that listen to new {@link MessageExchange} on the
     * specified channel. Exchanges are ignored depending on the specified
     * exchange-status to ignore.
     * 
     * @param component
     * @param jbiPool
     * @param exchange
     */
    public MessageExchangeProcessor(final AbstractComponent component,
            final AbstractJBIListener jbiListener, final Logger logger) {
        this.component = component;
        this.jbiListener = jbiListener;
        this.logger = logger;

        // get a direct reference to the required component attributes
        this.deliveryChannel = component.getChannel();
        this.suManager = (AbstractServiceUnitManager) component.getServiceUnitManager();
        this.asyncManager = component.getAsyncManager();
        // FIXME: Bug as the generated class of the Fractal binding to the
        // transaction
        // manager is not in the classloader of the component when this latest
        // is isolated. It is not possible to cast it in the component!
        this.transactionManager = component.getContext().getTransactionManager();
        this.componentName = component.getContext().getComponentName();
        this.ignoreStatus = component.getComponentConfiguration().getIgnoredStatus();
    }

    /**
     * Set the Exchange to process.
     * 
     * @param exchange
     */
    public void setMessageExchange(final ExchangeImpl exchange) {
        this.exchange = exchange;

    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Runnable#run()
     */
    public void run() {
        if (!this.isIgnored(this.exchange)) {
            try {
                this.process(this.exchange);
            } catch (final Exception e) {
                this.logger.log(Level.SEVERE, e.getMessage(), e);
                // Return an error acknowledgment only if the exchange is
                // not terminated
                if (this.exchange.isActiveStatus()) {
                    try {
                        // The exception must be managed according to the MEP
                        if (MEPConstants.IN_OUT_PATTERN.equals(this.exchange.getPattern())) {
                            // a fault must be returned
                            this.exchange.setFault(e);
                        } else {
                            // For the other MEPs, an error could be returned
                            this.exchange.setError(new MessagingException(ExceptionUtil
                                    .getExtendedMessage(e)));
                        }

                        this.deliveryChannel.send(this.exchange.getMessageExchange());
                    } catch (final MessagingException e1) {
                        this.logger.log(Level.SEVERE,
                                "Cannot send back an error acknowledgement or a Fault", e);
                    }
                }
            }
        }
    }

    /**
     * Clean the Message Exchange Processor. It cleans all the resources used
     * for one Message Processing.
     */
    protected final void clean() {
        // reset the JBI Listener instance
        this.jbiListener.setConsumes(null);
        this.jbiListener.setProvides(null);
        this.jbiListener.setExtensions(null);
        this.jbiListener.setCurrentTransaction(null);
        this.exchange = null;
    }

    /**
     * Launch the component processing of the given message exchange. If the
     * endpoint of the exchange is directly managed by the targeted component,<br />
     * no Provides exist: the component fully manages its kind of message.<br />
     * So the MessageInterceptorManager isn't called and no ExchangeProperties
     * are set.
     * 
     * @param exchange
     *            The message exchange to process. Must be non null
     * @throws InterruptedException
     * @throws MessagingException
     *             A problem occurred during manipulation of the exchange
     */
    private void process(final Exchange exchange) throws MessagingException {
        this.setCurrentTransaction();
        if (exchange.isProviderRole()) {
            try {
                this.processAsProvider(exchange);
            } catch (RuntimeException e) {
                if (exchange.isProviderRole()) {
                    throw e;
                }
            } catch (MessagingException e) {
                if (exchange.isProviderRole()) {
                    throw e;
                }
            }
        } else {
            try {
                this.processAsConsumer(exchange);
            } catch (RuntimeException e) {
                if (exchange.isConsumerRole()) {
                    throw e;
                }
            } catch (MessagingException e) {
                if (exchange.isConsumerRole()) {
                    throw e;
                }
            }
        }
    }

    /**
     * Process the exchange as service consumer
     * 
     * @param exchange
     * @throws MessagingException
     */
    private final void processAsConsumer(final Exchange exchange) throws MessagingException {
        AsyncContext asyncContext = null;

        this.logger.finest("Process an exchange as CONSUMER with id : " + exchange.getExchangeId());
        final Consumes consumes = this.getConsumesFromExchange(exchange);
        if (consumes != null) {
            this.jbiListener.setConsumes(consumes);
            this.jbiListener.setExtensions(this.getExtensions(consumes));
            // TODO: intercept even the asynchronous exchanges?
            try {
                this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, consumes,
                        InterceptorPhase.SEND_RESPONSE);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        }
        if (this.exchange.getProperty(AsyncMessageManager.ASYNC_MESSAGE_PROPERTY) != null) {
            // Asynchronous message handling for sub-consumer
            asyncContext = this.asyncManager.getAsyncContext(this.exchange.getExchangeId());
            if (asyncContext == null && exchange.isActiveStatus()) {
                exchange
                        .setError(new Exception(
                                "A response message do not match any Asynchronous Context. Its bound Asynchronous Context has probably already expired!"));
                this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
            }
        } else if (this.exchange.getProperty(AsyncMessageManager.ASYNC_ORIGINAL_MESSAGE_PROPERTY) != null) {
            // Asynchronous message handling for original consumer
            asyncContext = this.asyncManager.removeAsyncContext(this.exchange.getExchangeId());
        } else {
            this.logger
                    .fine("The exchange '"
                            + exchange.getExchangeId()
                            + "' refers to an unknown consumer on the component : "
                            + this.componentName
                            + ". The listener will not have the consumes and the extensions for this exchange");
        }
        if (exchange.isInOnlyPattern()) {
            this.processInOnlyAsConsumer(exchange, asyncContext);
        } else if (exchange.isRobustInOnlyPattern()) {
            this.processRobustInOnlyAsConsumer(exchange, asyncContext);
        } else if (exchange.isInOutPattern()) {
            this.processInOutAsConsumer(exchange, asyncContext);
        } else if (exchange.isInOptionalOutPattern()) {
            this.processInOptionalOutAsConsumer(exchange, asyncContext);
        } else {
            throw new MessagingException("Message Exchange Pattern not recognized :"
                    + exchange.getPattern());
        }
    }

    /**
     * Process the exchange as service provider
     * 
     * @param exchange
     *            The ongoing exchange
     * @param asyncContext
     *            The asynchronous context
     * @throws MessagingException
     */
    private final void processAsProvider(final Exchange exchange) throws MessagingException {
        this.logger.finest("Process an exchange as PROVIDER with id : " + exchange.getExchangeId());

        final Provides provides = this.getProvidesFromExchange(exchange);
        if (provides == null) {
            if (this.component.getServiceDescription(exchange.getEndpoint()) == null) {
                throw new MessagingException("The endpoint '" + exchange.getEndpoint()
                        + "' is unknown on the component : " + this.componentName);
            }
            this.logger.fine("Process an exchange managed directly by the component");
        } else {
            this.jbiListener.setProvides(provides);
            this.jbiListener.setExtensions(this.getExtensions(provides));
            try {
                this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, provides,
                        InterceptorPhase.ACCEPT);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        }
        // it can be an original Message Exchange of an expired Asynchronous
        // Context
        AsyncContext asyncContext = null;
        if (this.exchange.getProperty(AsyncMessageManager.ASYNC_EXPIRED_MESSAGE_PROPERTY) != null) {
            asyncContext = this.asyncManager.removeAsyncContext(this.exchange.getExchangeId());
        }
        // put an id to correlate all the exchanges of a same technical flow

        if (exchange.getProperty(Message.CORRELATION_ID_KEY) == null) {
            // add a correlation id
            final String correlationId = IDGenerator.getInstance().getNewID();
            exchange.setProperty(Message.CORRELATION_ID_KEY, correlationId);
        }

        // TODO : be less intrusive and remove the 'Check Role trick'
        ((ExchangeImpl) exchange).setCheckRole(false);
        if (this.component.getNotificationBrokerController() != null
                && !UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            this.component.getNotificationBrokerController().getNotificationProducer()
                    .getNotifyController().process(InProducerComponentNotify.class, exchange);
        }
        ((ExchangeImpl) exchange).setCheckRole(true);

        if (exchange.isInOnlyPattern()) {
            this.processInOnlyAsProvider(exchange, asyncContext, provides);
        } else if (exchange.isRobustInOnlyPattern()) {
            this.processRobustInOnlyAsProvider(exchange, asyncContext, provides);
        } else if (exchange.isInOutPattern()) {
            this.processInOutAsProvider(exchange, asyncContext, provides);
        } else if (exchange.isInOptionalOutPattern()) {
            this.processInOptionalOutAsProvider(exchange, asyncContext, provides);
        } else {
            throw new MessagingException("Message Exchange Pattern not recognized :"
                    + exchange.getPattern());
        }
    }

    /**
     * Process "in-only" MessageExchange as Consumer
     * 
     * @param exchange
     * @param asyncContext
     *            The asynchronous context
     * @throws MessagingException
     */
    private final void processInOnlyAsConsumer(final Exchange exchange,
            final AsyncContext asyncContext) throws MessagingException {
        if (UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            if (asyncContext == null) {
                this.jbiListener.onNotificationMessage(exchange);
            } else {
                this.jbiListener.onAsyncNotificationMessage(exchange, asyncContext);
            }
        } else {
            if (asyncContext == null) {
                this.jbiListener.onJBIMessage(exchange);
            } else {
                this.jbiListener.onAsyncJBIMessage(exchange, asyncContext);
            }
        }
    }

    /**
     * Process "in-only" MessageExchange As Provider
     * 
     * @param exchange
     *            the MessageExchange
     * @param provides
     *            The provides
     * @throws MessagingException
     *             When errors occur during processing
     */
    private final void processInOnlyAsProvider(final Exchange exchange,
            final AsyncContext asyncContext, final Provides provides) throws MessagingException {
        boolean reply = false;
        if (UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            if (asyncContext == null) {
                reply = this.jbiListener.onNotificationMessage(exchange);
            } else {
                reply = this.jbiListener.onExpiredAsyncNotificationMessage(exchange, asyncContext);
            }
        } else {
            if (asyncContext == null) {
                reply = this.jbiListener.onJBIMessage(exchange);
            } else {
                reply = this.jbiListener.onExpiredAsyncJBIMessage(exchange, asyncContext);
            }
        }
        if (reply) {
            if (exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
            if (provides != null) {
                try {
                    this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, provides,
                            InterceptorPhase.ACCEPT_RESPONSE);
                } catch (final PEtALSCDKException e) {
                    throw new MessagingException(e);
                }
                this.setProperties(provides);
            }
            // TODO : be less intrusive
            if (this.component.getNotificationBrokerController() != null
                    && !UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
                this.component.getNotificationBrokerController().getNotificationProducer()
                        .getNotifyController().process(StatusProducerComponentNotify.class,
                                exchange);
            }
            this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
        }
    }

    /**
     * Process "in-optional-out" MessageExchange Consumer.
     * 
     * @param exchange
     * @throws MessagingException
     */
    private final void processInOptionalOutAsConsumer(final Exchange exchange,
            final AsyncContext asyncContext) throws MessagingException {
        boolean reply = false;
        boolean doneOrErrorStatusReceived = false;
        if (exchange.isDoneStatus() || exchange.isErrorStatus()) {
            doneOrErrorStatusReceived = true;
        }
        if (UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            if (asyncContext == null) {
                reply = this.jbiListener.onNotificationMessage(exchange);
            } else {
                reply = this.jbiListener.onAsyncNotificationMessage(exchange, asyncContext);
            }
        } else {
            if (asyncContext == null) {
                reply = this.jbiListener.onJBIMessage(exchange);
            } else {
                reply = this.jbiListener.onAsyncJBIMessage(exchange, asyncContext);
            }
        }
        if (reply && !doneOrErrorStatusReceived) {
            if (exchange.getFault() == null && exchange.isActiveStatus()) {
                // Set Status DONE if no Fault and ACTIVE Status
                exchange.setDoneStatus();
            }
            this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
        }
    }

    /**
     * Process "in-optional-out" MessageExchange as Provider *
     * 
     * @param exchange
     *            the MessageExchange
     * @param provides
     *            The provides
     * @throws MessagingException
     *             When errors occur during processing
     */
    private final void processInOptionalOutAsProvider(final Exchange exchange,
            final AsyncContext asyncContext, final Provides provides) throws MessagingException {
        boolean reply = false;
        boolean faultReceived = false;
        boolean doneOrErrorStatusReceived = false;
        if (exchange.isDoneStatus() || exchange.isErrorStatus()) {
            doneOrErrorStatusReceived = true;
        } else if (exchange.getFault() != null) {
            faultReceived = true;
        }
        if (UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            if (asyncContext == null) {
                reply = this.jbiListener.onNotificationMessage(exchange);
            } else {
                reply = this.jbiListener.onExpiredAsyncNotificationMessage(exchange, asyncContext);
            }
        } else {
            if (asyncContext == null) {
                reply = this.jbiListener.onJBIMessage(exchange);
            } else {
                reply = this.jbiListener.onExpiredAsyncJBIMessage(exchange, asyncContext);
            }
        }
        if (provides != null && exchange.isOutMessage()) {
            forwardMessageData(exchange, provides);
        }
        // Set Status DONE if no Fault or Out messages and ACTIVE Status
        if (reply) {
            if (!doneOrErrorStatusReceived) {
                if (exchange.isActiveStatus()
                        && (faultReceived || exchange.getFault() == null
                                && !((ExchangeImpl) exchange).isOutMessage())) {
                    exchange.setDoneStatus();
                }
            }
            if (provides != null) {
                try {
                    this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, provides,
                            InterceptorPhase.ACCEPT_RESPONSE);
                } catch (final PEtALSCDKException e) {
                    throw new MessagingException(e);
                }
                this.setProperties(provides);
            }
            if (!doneOrErrorStatusReceived) {
                // TODO : be less intrusive
                if (this.component.getNotificationBrokerController() != null
                        && !UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
                    if (exchange.isActiveStatus()) {
                        if (exchange.getFault() == null) {
                            this.component.getNotificationBrokerController()
                                    .getNotificationProducer().getNotifyController().process(
                                            OutProducerComponentNotify.class, exchange);
                        } else {
                            this.component.getNotificationBrokerController()
                                    .getNotificationProducer().getNotifyController().process(
                                            FaultProducerComponentNotify.class, exchange);
                        }
                    } else {
                        this.component.getNotificationBrokerController().getNotificationProducer()
                                .getNotifyController().process(StatusProducerComponentNotify.class,
                                        exchange);
                    }
                }
                this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
            } else {
                this.component.getNotificationBrokerController().getNotificationProducer()
                        .getNotifyController().process(StatusProducerComponentNotify.class,
                                exchange);
            }
        }
    }

    /**
     * Process "in-out" MessageExchange as Consumer.
     * 
     * @param exchange
     * @throws MessagingException
     */
    private final void processInOutAsConsumer(final Exchange exchange,
            final AsyncContext asyncContext) throws MessagingException {
        boolean reply = false;
        if (UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            if (asyncContext == null) {
                reply = this.jbiListener.onNotificationMessage(exchange);
            } else {
                reply = this.jbiListener.onAsyncNotificationMessage(exchange, asyncContext);
            }
        } else {
            if (asyncContext == null) {
                reply = this.jbiListener.onJBIMessage(exchange);
            } else {
                reply = this.jbiListener.onAsyncJBIMessage(exchange, asyncContext);
            }
        }
        if (reply) {
            if (exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
            this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
        }
    }

    /**
     * Process "in-out" MessageExchange as Provider.<br />
     * This pattern is special, as it is bound to the request/response mode in
     * WSDL, the provider MUST return a response, no acknowledgment is allowed
     * at this step of the pattern
     * 
     * @param exchange
     *            the MessageExchange
     * @param asyncContext
     *            The asynchronous context
     * @param provides
     *            The provides
     * @throws MessagingException
     *             When errors occur during processing
     * 
     */
    private final void processInOutAsProvider(final Exchange exchange,
            final AsyncContext asyncContext, final Provides provides) throws MessagingException {
        boolean reply = false;
        boolean doneOrErrorStatusReceived = false;
        try {
            if (exchange.isDoneStatus() || exchange.isErrorStatus()) {
                doneOrErrorStatusReceived = true;
            }
            if (UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
                if (asyncContext == null) {
                    reply = this.jbiListener.onNotificationMessage(exchange);
                } else {
                    reply = this.jbiListener.onExpiredAsyncNotificationMessage(exchange,
                            asyncContext);
                }
            } else {
                if (asyncContext == null) {
                    reply = this.jbiListener.onJBIMessage(exchange);
                } else {
                    reply = this.jbiListener.onExpiredAsyncJBIMessage(exchange, asyncContext);
                }
            }

            if (provides != null && exchange.isOutMessage()) {
                forwardMessageData(exchange, provides);
            }
        } catch (final Exception e) {
            this.logger.log(Level.WARNING,
                    "An unexpected exception occured during an exchange processing", e);
            // set Fault message if reply is allowed, an error has been raised
            // and the exchange isn't terminated
            if (exchange.getFault() == null) {
                exchange.setFault(new SOAP11FaultServerException(e.getMessage(),
                        Constants.Exception.NAMESPACE_URI.resolve("/" + this.componentName), e));
                reply = true;
            }
        }
        if (reply && !doneOrErrorStatusReceived) {
            // set Fault message if no Fault or Out message
            if (exchange.getFault() == null && !exchange.isOutMessage()) {
                // an error may have been set as in contrary of the status, it
                // is not forbidden to set an error on the exchange in this
                // transition of this pattern
                if (exchange.getError() != null) {
                    exchange.setFault(new SOAP11FaultServerException(exchange.getError()
                            .getMessage(), Constants.Exception.NAMESPACE_URI.resolve("/"
                            + this.componentName), exchange.getError()));
                } else {
                    exchange.setFault(new SOAP11FaultServerException(
                            "The Provider do not return any response",
                            Constants.Exception.NAMESPACE_URI.resolve("/" + this.componentName)));
                }
            }
            if (provides != null) {
                try {
                    this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, provides,
                            InterceptorPhase.ACCEPT_RESPONSE);
                } catch (final PEtALSCDKException e) {
                    throw new MessagingException(e);
                }
                this.setProperties(provides);
            }
            // TODO : be less intrusive
            if (this.component.getNotificationBrokerController() != null
                    && !UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
                if (exchange.getFault() == null) {
                    this.component.getNotificationBrokerController().getNotificationProducer()
                            .getNotifyController().process(OutProducerComponentNotify.class,
                                    exchange);
                } else {
                    this.component.getNotificationBrokerController().getNotificationProducer()
                            .getNotifyController().process(FaultProducerComponentNotify.class,
                                    exchange);
                }
            }
            this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
        }
    }

    /**
     * Process "In-only" MessageExchange as consumer
     * 
     * @param exchange
     * @param asyncContext
     *            The asynchronous context
     * @throws MessagingException
     */
    private final void processRobustInOnlyAsConsumer(final Exchange exchange,
            final AsyncContext asyncContext) throws MessagingException {

        // This can be only an acknowledgment, according to the InOnly MEP.

        if (UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            if (asyncContext == null) {
                this.jbiListener.onNotificationMessage(exchange);
            } else {
                this.jbiListener.onAsyncNotificationMessage(exchange, asyncContext);
            }
        } else {
            if (asyncContext == null) {
                this.jbiListener.onJBIMessage(exchange);
            } else {
                this.jbiListener.onAsyncJBIMessage(exchange, asyncContext);
            }
        }
    }

    /**
     * Process "robust-in-only" MessageExchange as Provider
     * 
     * @param exchange
     *            the MessageExchange
     * @param provides
     *            The provides
     * @throws MessagingException
     *             When errors occur during processing
     * 
     */
    private final void processRobustInOnlyAsProvider(final Exchange exchange,
            final AsyncContext asyncContext, final Provides provides) throws MessagingException {
        boolean reply = false;
        boolean doneOrErrorStatusReceived = false;
        if (exchange.isDoneStatus() || exchange.isErrorStatus()) {
            doneOrErrorStatusReceived = true;
        }
        if (UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            if (asyncContext == null) {
                reply = this.jbiListener.onNotificationMessage(exchange);
            } else {
                reply = this.jbiListener.onExpiredAsyncNotificationMessage(exchange, asyncContext);
            }
        } else {
            if (asyncContext == null) {
                reply = this.jbiListener.onJBIMessage(exchange);
            } else {
                reply = this.jbiListener.onExpiredAsyncJBIMessage(exchange, asyncContext);
            }
        }
        if (reply && !doneOrErrorStatusReceived) {
            if (exchange.getFault() == null && exchange.isActiveStatus()) {
                // Set Status DONE if reply is allowed, no fault and ACTIVE
                // status
                exchange.setDoneStatus();
            }
            if (provides != null) {
                try {
                    this.jbiListener.getMessageInterceptorProcessor().intercept(exchange, provides,
                            InterceptorPhase.ACCEPT_RESPONSE);
                } catch (final PEtALSCDKException e) {
                    throw new MessagingException(e);
                }
                this.setProperties(provides);
            }
            // TODO : be less intrusive
            if (this.component.getNotificationBrokerController() != null
                    && !UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
                final AbstractProducerComponentNotify notify = null;
                if (exchange.isActiveStatus()) {
                    if (exchange.getFault() == null) {
                        this.component.getNotificationBrokerController().getNotificationProducer()
                                .getNotifyController().process(OutProducerComponentNotify.class,
                                        exchange);
                    } else {
                        this.component.getNotificationBrokerController().getNotificationProducer()
                                .getNotifyController().process(FaultProducerComponentNotify.class,
                                        exchange);
                    }
                } else {
                    this.component.getNotificationBrokerController().getNotificationProducer()
                            .getNotifyController().process(StatusProducerComponentNotify.class,
                                    exchange);
                }
            }
            this.deliveryChannel.send(((ExchangeImpl) exchange).getMessageExchange());
        }
    }

    /**
     * Get the extensions for the given provides/consumes.
     * 
     * @param service
     * @return The extensions
     */
    private final ConfigurationExtensions getExtensions(final Object service) {
        ConfigurationExtensions extensions = null;
        if (service == null) {
            return null;
        }
        final ServiceUnitDataHandler suDataHandler = this.suManager
                .getSUDataHandlerForService(service);
        if (suDataHandler != null) {
            extensions = suDataHandler.getConfigurationExtensions(service);
        }
        return extensions;
    }

    /**
     * Get the Consumes element for the given exchange.
     * 
     * @param exchange
     * @return {@code Consumes}
     */
    private final Consumes getConsumesFromExchange(final Exchange exchange) {
        final ServiceEndpoint endpoint = exchange.getEndpoint();
        return this.suManager.getConsumesFromDestination(endpoint.getEndpointName(), exchange
                .getService(), exchange.getInterfaceName());
    }

    /**
     * Get the Provides element for the given exchange.
     * 
     * @param exchange
     * @return {@code Provides}
     */
    private final Provides getProvidesFromExchange(final Exchange exchange) {
        final ServiceEndpoint ep = exchange.getEndpoint();
        final Provides provides = this.suManager.getProvidesFromEndpoint(ep);
        return provides;
    }

    /**
     * Set the current transaction.
     */
    private final void setCurrentTransaction() {
        if (this.exchange.isTransacted()) {
            // PEtALS Container
            Transaction currentTransaction = (Transaction) this.exchange
                    .getProperty(MessageExchange.JTA_TRANSACTION_PROPERTY_NAME);

            // Potential other JBI container
            if (currentTransaction == null) {
                try {
                    // FIXME: Bug as the generated class of the Fractal binding
                    // to the transaction manager is not in the classloader of
                    // the component when this latest is isolated. It is not
                    // possible to cast it in the component!
                    currentTransaction = ((TransactionManager) this.transactionManager)
                            .getTransaction();
                } catch (final SystemException e) {
                    this.logger.log(Level.WARNING, "Failed to retrieve Transaction for exchange '"
                            + this.exchange.getExchangeId() + "'", e);
                }
            }
            this.jbiListener.setCurrentTransaction(currentTransaction);
        }
    }

    /**
     * Set the configured Exchange and Message Properties.
     * 
     * @param exchangeProperties
     * @throws MessagingException
     */
    private final void setProperties(final Provides provides) throws MessagingException {
        final ExchangeProperties exchangeProperties = provides.getExchangeProperties();
        if (exchangeProperties != null) {
            for (final Param exchangeProperty : exchangeProperties.getExchangeProperty()) {
                this.exchange.setProperty(exchangeProperty.getName(), exchangeProperty.getValue());
            }
        }

        final MessageProperties messageProperties = provides.getMessageProperties();
        if (messageProperties != null) {
            for (final MessageParam messageProperty : messageProperties.getMessageProperty()) {
                final NormalizedMessage message = this.exchange.getMessageExchange().getMessage(
                        messageProperty.getMessage().name());
                message.setProperty(messageProperty.getName(), messageProperty.getValue());
            }
        }
    }

    /**
     * Depending on the IgnoredStatus, tells if the exchange has to be ignored.
     * If the exchange is null, return true.
     * 
     * @param exchange
     * @return
     */
    private final boolean isIgnored(final Exchange exchange) {
        boolean result = false;

        if (exchange.isDoneStatus()) {
            result = IgnoreStatus.DONE_IGNORED == this.ignoreStatus
                    || IgnoreStatus.DONE_AND_ERROR_IGNORED == this.ignoreStatus;
        } else if (exchange.isErrorStatus()) {
            result = IgnoreStatus.ERROR_IGNORED == this.ignoreStatus
                    || IgnoreStatus.DONE_AND_ERROR_IGNORED == this.ignoreStatus;
        }
        return result;
    }

    /**
     * Forward the data from the In to the Out message.
     * 
     * @param exchange
     * @param provides
     * @throws MessagingException
     */
    private static final void forwardMessageData(final Exchange exchange, final Provides provides)
            throws MessagingException {

        if (provides == null) {
            return;
        }

        // forward security subject
        if (provides.isForwardSecuritySubject() != null
                && provides.isForwardSecuritySubject().booleanValue()) {
            if (exchange.getInSubject() != null && exchange.getOutSubject() == null) {
                exchange.setOutSubject(exchange.getInSubject());
            }
        }

        // forward message properties
        if (provides.isForwardMessageProperties() != null
                && provides.isForwardMessageProperties().booleanValue()) {
            for (final String propertyName : exchange.getInMessagePropertyNames()) {
                if (exchange.getOutMessageProperty(propertyName) == null) {
                    exchange.setOutMessageProperty(propertyName, exchange
                            .getInMessageProperty(propertyName));
                }
            }
        }

        // forward message attachment
        if (provides.isForwardAttachments() != null
                && provides.isForwardAttachments().booleanValue()) {
            final Set<String> attachmentNames = exchange.getInMessageAttachmentNames();
            for (final String attachmentName : attachmentNames) {
                if (exchange.getOutMessageAttachment(attachmentName) == null) {
                    exchange.setOutMessageAttachment(attachmentName, exchange
                            .getInMessageAttachment(attachmentName));
                }
            }
        }
    }

}
