/**
 * PETALS: PETALS Services Platform Copyright (C) 2009 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 */

package org.ow2.petals.component.framework.util;

import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

/**
 * A task executor which launches atomic calls at a lowering rate until a call
 * succeed.
 * 
 * @author chamerling - eBM WebSourcing
 * 
 */
public class RetryTaskExecutor {

    /**
     * The executed task name
     */
    private String taskName;

    private Logger logger;

    /**
     * The initial delay before launching the first task call
     */
    private final int minDelay;

    /**
     * The maximum delay between calls. This maximum value is the upper value
     * and two consecutive calls will not exceed this value.
     */
    private final int maxDelay;

    /**
     * The factor to compute next delay. When a call fail, the next call will be
     * called after current (delay * delayFactor)
     */
    private final int delayFactor;

    /**
     * The global timeunit of minDelay and maxDelays values
     */
    private final TimeUnit timeUnit;

    private int innerDelay;

    /**
     * Used to limit the number of calls. Even if we define a maximum delay, the
     * calls will be discarded and the current executor will stop when the total
     * number of calls is equal to the retires value. A value <= 0 means no
     * limit.
     */
    private final int retries;

    private ScheduledExecutorService executor;

    private volatile boolean running = true;

    public RetryTaskExecutor(String taskName, Logger logger, int minDelay, int maxDelay,
            int delayFactor, TimeUnit timeUnit, int retries) {
        super();
        if (maxDelay < minDelay) {
            throw new IllegalArgumentException("Max delay can not be less than min delay");
        }
        if (delayFactor < 1) {
            throw new IllegalArgumentException("Factor can not be less than 1");
        }
        if (timeUnit == null) {
            throw new IllegalArgumentException("TimeUnit can not be null");
        }
        if (!StringHelper.isNullOrEmpty(taskName)) {
            this.taskName = taskName;
        } else {
            this.taskName = "Undefined task";
        }

        this.minDelay = minDelay;
        this.maxDelay = maxDelay;
        this.timeUnit = timeUnit;
        this.delayFactor = delayFactor;
        this.retries = retries;
        this.logger = logger;
    }

    public <T> T tryExecute(final Callable<T> command) throws Exception {
        T result = null;
        this.logger.fine("Start execution of task: " + taskName);
        // Schedule a Future task to call the command after delay milliseconds.
        this.innerDelay = this.minDelay;
        this.executor = Executors.newScheduledThreadPool(1, new CustomizableThreadFactory(taskName
                + " - Retry Thread #"));
        long i = 0;
        while (this.running) {
            final long retry = i++;
            ScheduledFuture<T> future = this.executor.schedule(new Callable<T>() {
                @Override
                public T call() throws Exception {
                    return command.call();
                }
            }, this.innerDelay, this.timeUnit);

            try {
                // wait for result...
                result = future.get();
            } catch (ExecutionException e) {
                Throwable cause = e.getCause();

                if (retry == 1) {
                    // The retry process start, log it
                    this.logger.warning("Can't execute a provided task '" + this.taskName
                            + "', cause : " + cause.getMessage());
                    if (this.retries > 0) {
                        this.logger.warning("Retry the execution of this task " + this.retries
                                + " times before complete failure.");
                    } else {
                        this.logger.warning("Retry the execution of this task indefinitely");
                    }
                }

                if ((this.retries > 0) && (retry == this.retries - 1)) {
                    this.executor.shutdownNow();
                    throw new Exception("Can not get a valid response in " + this.retries
                            + " tries...", cause);
                }

                // get the next delay...
                if ((this.innerDelay < this.maxDelay)
                        && ((this.innerDelay * this.delayFactor) < this.maxDelay)) {
                    this.innerDelay = this.innerDelay * this.delayFactor;
                } else {
                    this.innerDelay = this.maxDelay;
                }
            }

            // end of loop?
            if (result != null) {
                this.logger.fine("Task '" + this.taskName + "' succeed.");
                this.running = false;
            } else if ((this.retries > 0) && (retry == this.retries - 1)) {
                this.logger.warning("Retry proccess for task '" + this.taskName
                        + "' stops without result.");
                this.running = false;
            }

        }
        this.executor.shutdownNow();
        return result;
    }

    /**
     * Get the actual delay value (updated on each call)
     * 
     * @return
     */
    public int getActualDelay() {
        return this.innerDelay;
    }

    /**
     * Cancel the call ie interupts pending calls...
     */
    public void cancel() {
        this.running = false;
        if (this.executor != null) {
            this.executor.shutdownNow();
        }
    }
}
